/* Auswertung zweier DHT22 Temperatur- und Feuchtesensoren
Die Daten werden an ein LCD, USB, UART und ein Funkmodul RF02 ausgegeben.
Noch offen: Einstellung der Sendeleistung mit RF02init verknüpfen
DIESE DATEI MUSS IM ZEICHENSATZ UTF-8 GESPEICHERT SEIN!

Controller: AT90USB82, AT90USB162, ATmega8U2, ATmega16U2, ATmega32U2
(Achtung: Trotz „mega“ haben diese Controller keinen MUL-Befehl!
 Sowie keinen A/D-Wandler.)

Hardware: CPU-Portbelegung
PB0 14	(SS)	Funkmodul RF02 SDI
PB1 15	(SCLK)	Funkmodul RF02 SCK
PB2 16	(MOSI)	Funkmodul RF02 CS
PB3 17	(MISO)	Funkmodul RF02 IRQ
PB4 18	(T1)	DIL-Schalter nach L (Sendekanal)
PB5 19	-	DIL-Schalter nach L
PB6 20	-	DIL-Schalter nach L
PB7 21	(OC0A)	DIL-Schalter nach L

PC0  2	XTAL2	Quarz 16 MHz
PC1 24	!Reset	L = Rücksprung zum Urlader „ubaboot“
PC2  5	-	LCD RS (0 = Kommando, 1 = Daten)
PC4 26	-	LCD D4
PC5 25	(OC1B)	LCD D5
PC6 23	(OC1A)	LCD D6
PC7 22	(ICP1)	LCD D7

PD0  6	(OC0B)	Funktionstaste (ungenutzt)
PD1  7	(AIN)	(Interface Pin 4 = RS485-Richtungsumschaltung?)
PD2  8	RxD	RS232-Schnittstelle
PD3  9	TxD	RS232-Schnittstelle
PD4 10	-	Temperatur/Feuchtesensor DHT22 #0
PD5 11	(XCK)	Temperatur/Feuchtesensor DHT22 #1
PD6 12	(!RTS)	Power-Jumper: L = reduzierte Sendeleistung
PD7 13	(!CTS)	LCD E (H-aktiv)
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>		// nur _delay_us!!
#undef _delay_ms
#include "usb.h"
#include <avr/signature.h>
#include <avr/fuse.h>
FUSES={
 0b11011110,
 0b11011001,
 0b11111000,
};

static unsigned set_frequency(byte i) {	// Liefert Frequenz in Zehntel Megahertz
 // Frequenzdifferenz zu 860 MHz in Zehntel MHz (ab 90 unzulässig lt. Allgemeinzuteilung)
 static const byte frequency[] PROGMEM={80,40,50,70,90,110,130,150,170,190};
 return pgm_read_byte(frequency+i)+8600;
}

/*****************************************************************************************************************/
static byte TX NOINIT;		// Sender 1...10 -> (0...9)
static char sTemp[6] NOINIT;
static char sHumi[6] NOINIT;
static bool Power NOINIT;	// Ausgangsleistung

enum{iNaN=-0x8000};

// Dezimalzahl mit einer Nachkommastelle ausgeben
static void itoa_komma_1(char*s,int v) {
 if (v==iNaN) strcpy_P(s,PSTR("NaN"));
 else{
  s+=sprintf_P(s,PSTR("%02d"),v);	// eigenes sprintf() würde "%,1d" verarbeiten
  s[1]=0; *s=s[-1]; s[-1]=',';		// letzte Stelle hinterschieben
 }
}

// Vorzeichen-Betrags-Darstellung in Zweierkomplement umwandeln
static int vb(int v) {
 if (v&0x8000) v=-(v&0x7FFF);
 return v;	
}

// Hier DPCs hineinpacken!
static void idle(void) {
 usbPoll();
}

static void wait(unsigned ticks) {
 unsigned tic = TCNT1;
 do idle();
 while (TCNT1-tic < ticks);
}

inline static void wait_ms(float ms) {
 wait(ms/1000*F_CPU/1024);
}

// liest die Daten vom Temperatur- und Feuchtesensor
// Muss an PortD angeschlossen sein.
static void read_dht22_dat(byte pinmask){
 byte dht22_dat[4];
 byte laststate = pinmask;
 byte j = 0, b, sum4=0;
 PORTD |= pinmask;
 DDRD  |= pinmask;	// DHT auf H
 wait_ms(10);		// 10ms warten 
 PORTD &=~pinmask;	// DHT auf L
 wait_ms(18);		// 18ms warten 
 PORTD |= pinmask;	// DHT auf H
 _delay_us(40);		// 40us warten
 DDRD  &=~pinmask;	// DHT auf h: Lesemodus
  // Änderungen erkennen und Daten lesen
 int t=iNaN, h=iNaN;
 for (byte i=0; i< 85; i++) {
  byte counter = 0;
  while ((PIND&pinmask) == laststate) { // letzten Status mit Zustand des DHTPIN vergleichen
   _delay_us(1); // 1us warten 
   if (!++counter) goto err;
  }
  laststate = PIND&pinmask;	// Zustand DHTPIN lesen
// die ersten 3 Übergänge ignorieren
  if (i >= 4 && !(i&1)) {
      // jedes Bit in die Speicherbytes schieben
   b<<=1;
   if (counter > 16) b|=1;
   if (!(++j&~7)) {	// j (Bitnummer) durch 8 teilbar?
    if (j>=40) {
     if (b!=sum4) goto err;	// Kontrolle der Prüfsumme bei Bit 40
    }else sum4+=dht22_dat[j>>3]=b;	// summieren und abspeichern bei Bit 8,16,24,32
   }
  }
 }
 h=dht22_dat[0]<<8|dht22_dat[1];
 t=vb(dht22_dat[2]<<8|dht22_dat[3]);
err:
 itoa_komma_1(sTemp,t);
 itoa_komma_1(sHumi,h);
}

// ************************** Displayansteuerung ***************************
/* Version für HD44780-kompatible LCD 16×2 oder 20×2, Katakana-Zeichensatz
 * Das LCD ist folgendermaßen angeschlossen:
 * PD7     = Enable (H-aktiv)
 * PC2     = RS (0 = Kommando, 1 = Daten)
 * PC4-PC7 = D4-D7
 * R/W ist auf W festgelegt
 */

// H-Impuls 1 ms auf Enable-Leitung ausgeben
static void lcd_flash_e(void){
 PORTD |= 1<<7;		// Enable = HIGH
 wait_ms(1);		// kurz warten, damit Flanke entsteht
 PORTD &= ~(1<<7);	// Enable = LOW
 wait_ms(1);		// Warten, bis der durch Enable angelegte Befehl hoffentlich verarbeitet wurde
}

// Kommando oder Zeichen an Display senden 
// rs=false => Kommando, rs=true => Zeichen
static void lcd_write(uint8_t data, bool rs) {
 uint8_t pc=PORTC&0b00001011;
 if (rs) pc|=1<<2; 	// write data (RS=1)
 PORTC = pc | data&0xF0;// output high nibble
 lcd_flash_e();    
 PORTC = pc | data<<4;	// output low nibble
 lcd_flash_e();
}

// Display löschen
static void lcd_cls(void){
 lcd_write(0x00,0);   	// B 0000 0000 => Clear
 wait_ms(4);		// dauert eine Weile
 lcd_write(0x01,0);   	// B 0000 0001 => Cursor Home
 wait_ms(4);		// dauert eine Weile, Wert ausprobiert
}

// Zeichenausgabe
static void lcd_writechar(char zeichen) {lcd_write(zeichen,1);}

// Cursor positionieren
static void lcd_gotopos(byte zeile, byte spalte) {
// spalte+=(zeile>>1)*20;	// Nur für vierzeilige LCD
 if (zeile&1) spalte+=64;
 lcd_write(0x80|spalte,0);	// DD-RAM-Adresse setzen
}

// Zuerst den 8-Bit-Modus aktivieren, ist wichtig, falls LCD schon im 4-Bit Modus war und dann 
// nach einem Programm-Reset vergeblich versuchen würde, den 4-Bit Modus erneut zu aktivieren.
// Dann in 4-Bit Modus umschalten.
static void lcd_set4bit(void) {
	// 8-Bit Modus aktivieren (B 0011 0000 = H 30, High Nibble => B 0011 = H 03
 PORTC&=0b00001011;	// alles 0
 lcd_flash_e();    	// Datenübernahme
 wait_ms(5);		// sicherheitshalber warten da wir BUSY nicht überprüfen können 
	// Das ganze 2x wiederholen:
 lcd_flash_e();    
 _delay_us(200);			
 lcd_flash_e();    
 _delay_us(200);			
 lcd_write(0x2,0);	// B 0000 0010 => in 4-Bit-Modus umschalten 
 lcd_write(0x2,0);
 _delay_us(200);				
}

// Display initialisieren. einmal als erstes aufrufen
static void lcd_ini(void){
 wait_ms(50);		// richtig lange warten bis LCD wirklich bereit ist!
 lcd_set4bit();		// ab jetzt 4-Bit-Modus
 lcd_write(0x28,0);	// B 0010 1000 => Function Set: 4Bit (kein 8-Bit Bus), zweizeiliges Display, 5x7 Dots/Zeichen (kein 5x10), RE-Bit aus
 lcd_write(0x0C,0);	// B 0000 1000 => Display On/Off: Display ein, Cursor aus, Blinken aus
 lcd_write(0x06,0);	// B 0000 0110 => Entry Mode Set: DD-RAM autom. inkr., Cursor bewegen
 lcd_cls();		// Display löschen
}

// *********************** Ende Displayansteuerung *************************
//////////////////////////////// Funkmodul TX //////////////////////////////

#define RF_PORT	PORTB
#define RF_DDR	DDRB
#define RF_PIN	PINB// Modulbeschaltung:
#define SDI		0	// SDI,  -> RF02   --PB0
#define SCK		1	// SCK,  -> RF02   --PB1
#define CS		2	// nSEL, -> RF02   --PB2
#define IRQ		3	// nIRQ, <- RF02   --PB3
					// FSK: Pullup oder Pulldown
/*
433Band: Fc=430+F*0.0025 MHz => 433,05MHz...434,79 MHz
868Band: Fc=860+F*0.0050 MHz => 862,00MHz...868,00 MHz
https://www.mikrocontroller.net/articles/Allgemeinzuteilung
*/

// Hier: Zehntel MHz
inline unsigned RF02FREQ_860(unsigned f_MHz10) {return (f_MHz10-8600)*20;}	// für 860 MHz Modul
inline unsigned RF02FREQ_433(unsigned f_MHz10) {return (f_MHz10-4300)*40;}	// für 433 MHz Modul

/* Rechenbeispiele
z.B. RF02FREQ_860 = (868MHZ-860.0MHz)/0,0050) = 1600
z.B. RF02FREQ_433 = (434MHZ-430.0MHz)/0,0025) = 1600
*/
// Konstruktionsschwäche: Warum wurde nicht die eingebaute SPI-Hardware verwendet?
static void rf02_trans(unsigned wert) {
 byte i;
 RF_PORT&=~(1<<CS);
 for (i=0; i<16; i++) {
  if (wert&0x8000) RF_PORT|=1<<SDI;
  else RF_PORT&=~(1<<SDI);
  RF_PORT|=1<<SCK;
  wert<<=1;
  _delay_us(0.3);
  RF_PORT&=~(1<<SCK);
 }
 RF_PORT|=1<<CS;
}

// 1. Configuration Setting Command gem. Datenblatt
static void rf02_setmodfreq(byte bandwidth) {
//rf02_trans(0x8F80|bandwidth&7);  // 433MHz
 rf02_trans(0x9780|bandwidth&7); // 868MHz
//rf02_trans(0x8F70);
}

// 3. Frequency Setting Command gem. Datenblatt
static void rf02_setfreq(unsigned freq) {
 if (freq<96) freq=96;		// 430,2400MHz
 else if (freq>3903) freq=3903;	// 439,7575MHz
 rf02_trans(0xA000|freq);
//rf02_trans(0xA640); //0xC0
}

// 5. Power Setting Command gem. Datenblatt 
/*
p2 p1 p0 Output Power[dB]
0  0  0 		 0
0  0  1 		-3
0  1  0 		-6
0  1  1 		-9
1  0  0 		-12
1  0  1 		-15
1  1  0 		-18
1  1  1 		-21

bit 7 6 5 4 3  2  1  0 POR
	1 0 1 1 0 p2 p1 p0 B0h
The output power is given in the table as relative to the maximum available power,
which depends on the actual antenna impedance.
*/

static void rf02_setpower(byte power) {
 rf02_trans(0xB000|(power&7)<<8);
}

// 11. PLL Setting Command und 4. Data Rate Command gem. Datenblatt
static void rf02_setdata_rate(void) {
/*	PLL Setting Command
	
	D240h  19.2 ; Max data rate [kbps]
	D2C0h  38.4 ; Max data rate [kbps]	gewählt
	D200h  68.9 ; Max data rate [kbps]
	D280h 115.2 ; Max data rate [kbps]
*/	
 rf02_trans(0xD2C0);		// gewählt => 0b11010010 11000000
//  Data Rate Command
 rf02_trans(0xC806);		// gewählt => 0b11001000 00000110
}

static void rf02_shiftout(byte wert) {
 byte j;
 for (j=0; j<8; j++) {
  while (RF_PIN&(1<<IRQ));	// warte bis IRQ = L
  while (!(RF_PIN&(1<<IRQ)));	// warte bis IRQ = H	(externer Takt!)
  if (wert&128)	RF_PORT|=1<<SDI;		//100101000
  else RF_PORT&=~(1<<SDI);
  wert<<=1;
 }
}

static void rf02_txdata(const char*data, byte len) {
 byte i,wert=0xC6;	//1100 0110 ;9. Data Transmit Command
 RF_PORT&=~(1<<CS);	//nSel
 for (i=0; i<8; i++) {
  if (wert&0x80) RF_PORT|=1<<SDI;
  else RF_PORT&=~(1<<SDI);
  RF_PORT|=1<<SCK;
  wert<<=1;
  _delay_us(0.2);
  RF_PORT&=~(1<<SCK);
 }
 rf02_shiftout(0xAA);//10101010
 rf02_shiftout(0xAA);
 rf02_shiftout(0xAA);
 rf02_shiftout(0x2D);//00101101
 rf02_shiftout(0xD4);//11010100
 for (i=0; i<len; i++) rf02_shiftout(*data++);
 RF_PORT|=1<<CS;
 while (RF_PIN&(1<<IRQ));	// wait until transfer done
 rf02_trans(0xC464);		// TX off after 10us ;7. Sleep Command
}

static void rf02_init(void) {
 RF_PORT=(1<<CS);
 RF_DDR=(1<<SDI)|(1<<SCK)|(1<<CS);

 wait_ms(150);			// wait until POR done
 rf02_trans(0xC0E0);		// power settings				; 2. Power Management Command gem.Datenblatt
 rf02_trans(0x8A75);		// fsk in rfm02 = afc in rf12			; ??? vermutlich verstümmeltes "Configuration Setting Command"
//rf02_trans(0x80C7);
 rf02_trans(0xC2A0);		// enable tx sync bit, disable low bat detector	; 6. Low Battery Detector and TX Bit Synchronization Command gem. Datenblatt
 rf02_setfreq(RF02FREQ_860(set_frequency(TX)));// Sendefrequenz einstellen	; 3. Frequency Setting Command gem. Datenblatt (Mittenfrequenz)
 rf02_setpower(Power?0:1);	// Datenblatt  0db,-3db				; 5. Power Setting Command gem. Datenblatt 
 rf02_setmodfreq(3);		// 120kHz Frequenzshift 			; 1. Configuration Setting Command gem. Datenblatt
 rf02_setdata_rate();		// 11. PLL Setting Command und 
}				// 4. Data Rate Command gem. Datenblatt wurde festgelegt


enum{
 PRINT_LCD = 1,
 PRINT_SERIAL = 2,
 PRINT_USB = 4,
 PRINT_AIR = 8,
 PRINT_ALL = 7,	// alles außer Luft
};

static byte print_target = PRINT_ALL;

static void both_putchar(char c) {
 if (print_target&PRINT_SERIAL) {
  while (!(UCSR1A&1<<UDRE1)) idle();
  UDR1 = c;
 }
 if (print_target&PRINT_USB) usb_putchar(c);
}

// für printf-Funktion
static int all_putchar(char c, FILE *stream) {
 if (print_target&PRINT_AIR) {
  static char TX_buffer[20];
  static byte TX_index;
  TX_buffer[TX_index++]=c;
  if (c=='\n') {
   rf02_txdata(TX_buffer,TX_index);
   TX_index=0;
  }
 }
 if (print_target&PRINT_LCD) {
  byte b=c;
  if (c<0) {
   static unsigned widechar NOINIT;	// Unterstützung für UTF8 in BMP0
   unsigned wc;
   if (c&0x40) {		// Lead-Byte?
    if (c&0x20) wc = c|0x10;	// 3-Byte-Kette?
    else wc = c&0x1F | 1<<10;	// Fertig-Bit passend setzen
   }else{			// Trail-Byte
    wc = widechar;
    byte full = wc>>8 & 1<<2;	// merkt sich das letzte herausgeschobene Bit
    wc = wc<<6 | c&0x3F;
    if (full) {
     switch (wc) {
      case L'°': b=0xDF; break;
      case L'ä': b=0xE1; break;
      case L'ö': b=0xEF; break;
      case L'ü': b=0xF5; break;
      case L'ß': b=0xE2; break;
      case L'Ω': b=0xF4; break;
      case L'¥': b=0x5C; break;	//  HD44780 hat kein '\'
      case L'→': b=0x7E; break;	// 		kein '~'
      case L'←': b=0x7F; break;
// Zeichen mit UTF-8-Erscheinungsbild können hier per programmierbare Zeichen gemappt werden.
// Etwas anderes als die Katakana-Zeichentabelle habe ich noch nie zu Gesicht bekommen.
//    default: assert(0); // Dieser Quelltext sollte gefälligst keine unkonvertierbaren Zeichen enthalten
     }
     lcd_writechar(b);
     wc=0;
    }
   }
   widechar = wc;	// rückschreiben
  }else switch (c) {
   case '\n': b=' '; break;	// TODO: Rest der Zeile löschen und neue Zeile
// case 'g':	// Die Zeichenbilder in 5×9-Matix werden auf der gängigen 
// case 'j':	// 5×8-Matrix unbrauchbar dargestellt. Ausprobieren!
   case 'p':
   case 'q': b|=0x80; break;	// konvertieren zu Kodes mit Unterlänge
// Zeichenbilder für '~' und '\' können hier gemappt werden.
  }
  lcd_writechar(b);	// suppress non-printable characters on LCD
 }
 if (c=='\n') both_putchar('\r');
 both_putchar(c);
 return 0;
}

// Diese Funktion blockiert!!
static int uart_getchar(FILE *stream) {
 while (!(UCSR1A&1<<RXC1)) idle();
 return UDR1;
}

static FILE mystream = FDEV_SETUP_STREAM(all_putchar, uart_getchar, _FDEV_SETUP_RW);

static void InitUART(void) {
// 1.Baudrate (9600,8,n,1)
 UBRR1 = F_CPU/16/9600-1;		// ( Baud, 8Bit, 1Stopbit)
// 2.Receiver freischalten, Int. Receiver freischalten, Transmitter freischalten
 UCSR1B = 1<<RXEN1 | 1<<RXCIE1 | 1<<TXEN1;
}

static void emit_sensor(unsigned char n) {
 print_target = PRINT_SERIAL | PRINT_USB;
 printf_P(PSTR("Temperatur%u: %s°C," " relative Luftfeuchte%u: %s%%\n"),n,sTemp,n,sHumi);
 print_target = PRINT_LCD;
 lcd_gotopos(n&1?0:1,0);
 printf_P(PSTR("%u.T:%s°C F:%s%%\n"),n,sTemp,sHumi);
 print_target = PRINT_AIR;
 printf_P(PSTR("Temp%u=%sC\n"),n,sTemp);
 printf_P(PSTR("relH%u=%s%%\n"),n,sHumi);
}

unsigned tick;
static void waitTick(unsigned ticks) {
 do idle();
 while (TCNT1-tick < ticks);	// ms warten
 tick += ticks;			// quarzgenau beim nächsten Warten
}

inline static void waitTick_ms(float ms) {
 waitTick(ms*F_CPU/1024/1000);
}

int __attribute__((OS_main)) main(void) {
 ACSR = 1<<ACD;		// Analog Comparator Disable
 PORTD = 0b01000111;	// pull up:PD6 -> Power Jumper PD2 -> RX-PIN; PD0 -> Funktionstaste
 DDRD  = 0b10001010;	// PD7 -> EN(Display); PD3 -> TX-PIN, PD1 = Senderichtung?
 PORTB = 0b11110000;	// pull up: PB7...PB4 DIL Schalter
 DDRC  = 0b11110100;	// für Display alles Ausgänge LCD (D7...D4,RS), 4 Datenbits
 TCCR1B = 5;	// Vorteiler 1024
 Power = PIND&1<<6 ? true : false; // wenn Jumper gesetzt -3db Ausgangsleistung
 byte DIL = PINB>>4;	// DIL-Schalter abfragen
 if (DIL>9) DIL = 0;	// Sender 1...10 -> 0...9
 TX = DIL;

 InitUART();
// Datenstrom umlenken 
 stdout = stdin = &mystream; 

// LCD starten
 lcd_ini(); 
 lcd_cls();	

// TX Funkmodul starten
 rf02_init(); 

// ******** USB ***********
// initialize the USB, don't wait for the host to set configuration.
// If the µC is powered without a PC connected to the USB port,
// USB remains unconfigured, no pain.
 usbInit();
 sei();
//Do CPU sleep in Idle Mode. Currently, no sleep_cpu() is used.
 MCUCR = 0b00100000;		

// Anzeige der Einstellungen auf Display, seriell, und wenn vorhanden auf USB
//1.Zeile
 lcd_gotopos(0,0);
 print_target = PRINT_ALL;
 itoa_komma_1(sTemp,set_frequency(TX));
 printf_P(PSTR("Sender%u: %sMHz\n"),TX,sTemp); // Sendername,Frequenz
//2.Zeile
 lcd_gotopos(1,0);
 printf_P(PSTR("Power:%S USB:%S\n"),
	Power?PSTR("high"):PSTR("low "),
	usb_configured()?PSTR("on"):PSTR("off"));
 uint8_t count;
 tick = TCNT1;
 for (count=5; count>-1;count--){
  lcd_gotopos(1,19);
  print_target=PRINT_LCD; 
  printf_P(PSTR("%u"),count);	// down-Counter auf zeile 2 am Ende
  print_target=PRINT_SERIAL|PRINT_USB; 
  printf_P(PSTR("Start in %u Sekunden.\r"),count);
  usb_flush();
  waitTick_ms(1000);
 }
 printf("\n");	
 lcd_cls();	

// Hauptschleife
 for(;;) { 
// Sensor1
  read_dht22_dat(1<<4);
  emit_sensor(1+TX*2);
  waitTick_ms(1000);
// Sensor2
  read_dht22_dat(1<<5);
  emit_sensor(2+TX*2);
  waitTick_ms(1000);
 }
 return 0;
}
